#include <stddef.h> // offsetof
#include <memory.h>
#include <new.h>
#include <assert.h>
#include <angelscript.h>

#include "floatimage.h"
#include "application.h"


int CFloatImage::AddRef()
{
	return ++refCount;
}

int CFloatImage::Release()
{
	if( --refCount == 0 )
	{
		delete this;
		return 0;
	}

	return refCount;
}

CFloatImage::CFloatImage()
{
	refCount = 1;

	pixels = 0;
	width = 0;
	height = 0;
}

CFloatImage::~CFloatImage()
{
	if( pixels )
		delete[] pixels;
}

CFloatImage &CFloatImage::operator=(const CFloatImage &other)
{
	if( pixels ) delete[] pixels;

	pixels = new float[other.width*other.height*4];
	width = other.width;
	height = other.height;
	memcpy(pixels, other.pixels, other.width*other.height*4*4);

	return *this;
}

int CFloatImage::Create(UINT w, UINT h)
{
	if( w < 1 ) w = 1;
	if( h < 1 ) h = 1;

	pixels = new float[w*h*4];

	memset(pixels, 0, w*h*4*4);

	width = w;
	height = h;

	return 0;
}

float *CFloatImage::Alpha(UINT x, UINT y)
{
	if( x >= (unsigned)width || y >= (unsigned)height )
	{
		asIScriptContext *context = asGetActiveContext();

		context->SetException("Out of range");
		return 0;
	}

	return pixels + (y*width+x)*4 + 3;
}

float *CFloatImage::Red(UINT x, UINT y)
{
	if( x >= (unsigned)width || y >= (unsigned)height )
	{
		asIScriptContext *context = asGetActiveContext();

		context->SetException("Out of range");
		return 0;
	}

	return pixels + (y*width+x)*4 + 2;
}

float *CFloatImage::Green(UINT x, UINT y)
{
	if( x >= (unsigned)width || y >= (unsigned)height )
	{
		asIScriptContext *context = asGetActiveContext();

		context->SetException("Out of range");
		return 0;
	}

	return pixels + (y*width+x)*4 + 1;
}

float *CFloatImage::Blue(UINT x, UINT y)
{
	if( x >= (unsigned)width || y >= (unsigned)height )
	{
		asIScriptContext *context = asGetActiveContext();

		context->SetException("Out of range");
		return 0;
	}

	return pixels + (y*width+x)*4 + 0;
}

float *CFloatImage::Channel(UINT ch, UINT x, UINT y)
{
	if( x >= (unsigned)width || y >= (unsigned)height || ch >= 4 )
	{
		asIScriptContext *context = asGetActiveContext();

		context->SetException("Out of range");
		return 0;
	}

	return pixels + (y*width+x)*4 + ch;
}

void CFloatImage::Resize(UINT w, UINT h)
{
	if( pixels == 0 )
		return;

	if( w < 1 ) w = 1;
	if( h < 1 ) h = 1;

	ResizeWidthAvg(w);
	ResizeHeightAvg(h);
}

void CFloatImage::ResizeWidthAvg(UINT w)
{
	if( w == width )
		return;

	float *p = new float[w*height*4];

	memset(p, 0, w*height*4*4);

	if( w < width )
	{
		int d = (width<<16)/w;
		for( UINT y = 0; y < height; y++ )
		{
			float *pd = &p[y*w*4];
			float *ps = &pixels[y*width*4];

			unsigned int sx = 0;
			unsigned int ex = d;
			for( UINT x = 0; x < w; x++ )
			{
				// Fractional pixel at the start
				if( sx & 0xFFFF )
				{
					float f = float(0x10000 - sx & 0xFFFF)/0x10000;

					pd[0] += f*ps[0];
					pd[1] += f*ps[1];
					pd[2] += f*ps[2];
					pd[3] += f*ps[3];

					// next source pixel
					ps += 4;
					sx = (sx & 0xFFFF0000) + 0x10000;
				}

				// Add whole pixels
				while( sx < (ex & 0xFFFF0000) )
				{
					pd[0] += ps[0];
					pd[1] += ps[1];
					pd[2] += ps[2];
					pd[3] += ps[3];

					// next source pixel
					ps += 4;
					sx += 0x10000;
				}

				// Fractional pixel at the end
				if( ex & 0xFFFF )
				{
					float f = float(ex & 0xFFFF)/0x10000;

					pd[0] += f*ps[0];
					pd[1] += f*ps[1];
					pd[2] += f*ps[2];
					pd[3] += f*ps[3];
				}

				// normalize
				float f = 1/(float(d)/0x10000);
				pd[0] *= f;
				pd[1] *= f;
				pd[2] *= f;
				pd[3] *= f;

				// next destination pixel
				pd += 4;
				sx = ex;
				ex += d;
			}
		}
	}
	else
	{
		int d = (w<<16)/width;
		for( UINT y = 0; y < height; y++ )
		{
			float *pd = &p[y*w*4];
			float *ps = &pixels[y*width*4];

			unsigned int sx = 0;
			unsigned int ex = d;
			for( UINT x = 0; x < width; x++ )
			{
				// Fractional pixel at the start
				if( sx & 0xFFFF )
				{
					float f = float(0x10000 - sx & 0xFFFF)/0x10000;

					pd[0] += f*ps[0];
					pd[1] += f*ps[1];
					pd[2] += f*ps[2];
					pd[3] += f*ps[3];

					// next destination pixel
					pd += 4;
					sx = (sx & 0xFFFF0000) + 0x10000;
				}

				// Add whole pixels
				while( sx < (ex & 0xFFFF0000) )
				{
					pd[0] += ps[0];
					pd[1] += ps[1];
					pd[2] += ps[2];
					pd[3] += ps[3];

					// next destination pixel
					pd += 4;
					sx += 0x10000;
				}

				// Fractional pixel at the end
				if( ex & 0xFFFF )
				{
					float f = float(ex & 0xFFFF)/0x10000;

					pd[0] += f*ps[0];
					pd[1] += f*ps[1];
					pd[2] += f*ps[2];
					pd[3] += f*ps[3];
				}

				// next source pixel
				ps += 4;
				sx = ex;
				ex += d;
			}
		}
	}

	width = w;
	delete[] pixels;
	pixels = p;
}

void CFloatImage::ResizeHeightAvg(UINT h)
{
	if( h == height )
		return;
	
	// Resize height
	float *p = new float[width*h*4];

	memset(p, 0, width*h*4*4);

	if( h < height )
	{
		int d = (height<<16)/h;
		for( UINT x = 0; x < width; x++ )
		{
			float *pd = &p[x*4];
			float *ps = &pixels[x*4];

			unsigned int sy = 0;
			unsigned int ey = d;
			for( UINT y = 0; y < h; y++ )
			{
				// Fractional pixel at the start
				if( sy & 0xFFFF )
				{
					float f = float(0x10000 - sy & 0xFFFF)/0x10000;

					pd[0] += f*ps[0];
					pd[1] += f*ps[1];
					pd[2] += f*ps[2];
					pd[3] += f*ps[3];

					// next source pixel
					ps += 4*width;
					sy = (sy & 0xFFFF0000) + 0x10000;
				}

				// Add whole pixels
				while( sy < (ey & 0xFFFF0000) )
				{
					pd[0] += ps[0];
					pd[1] += ps[1];
					pd[2] += ps[2];
					pd[3] += ps[3];

					// next source pixel
					ps += 4*width;
					sy += 0x10000;
				}

				// Fractional pixel at the end
				if( ey & 0xFFFF )
				{
					float f = float(ey & 0xFFFF)/0x10000;

					pd[0] += f*ps[0];
					pd[1] += f*ps[1];
					pd[2] += f*ps[2];
					pd[3] += f*ps[3];
				}

				// normalize
				float f = 1/(float(d)/0x10000);
				pd[0] *= f;
				pd[1] *= f;
				pd[2] *= f;
				pd[3] *= f;

				// next destination pixel
				pd += 4*width;
				sy = ey;
				ey += d;
			}
		}
	}
	else
	{
		int d = (h<<16)/height;
		for( UINT x = 0; x < width; x++ )
		{
			float *pd = &p[x*4];
			float *ps = &pixels[x*4];

			unsigned int sy = 0;
			unsigned int ey = d;
			for( UINT y = 0; y < height; y++ )
			{
				// Fractional pixel at the start
				if( sy & 0xFFFF )
				{
					float f = float(0x10000 - sy & 0xFFFF)/0x10000;

					pd[0] += f*ps[0];
					pd[1] += f*ps[1];
					pd[2] += f*ps[2];
					pd[3] += f*ps[3];

					// next destination pixel
					pd += 4*width;
					sy = (sy & 0xFFFF0000) + 0x10000;
				}

				// Add whole pixels
				while( sy < (ey & 0xFFFF0000) )
				{
					pd[0] += ps[0];
					pd[1] += ps[1];
					pd[2] += ps[2];
					pd[3] += ps[3];

					// next destination pixel
					pd += 4*width;
					sy += 0x10000;
				}

				// Fractional pixel at the end
				if( ey & 0xFFFF )
				{
					float f = float(ey & 0xFFFF)/0x10000;

					pd[0] += f*ps[0];
					pd[1] += f*ps[1];
					pd[2] += f*ps[2];
					pd[3] += f*ps[3];
				}

				// next source pixel
				ps += 4*width;
				sy = ey;
				ey += d;
			}
		}
	}

	height = h;
	delete[] pixels;
	pixels = p;
}

void FloatImage_Construct(CFloatImage *o)
{
	new(o) CFloatImage();
	o->Create(1,1);
}

void FloatImage_Show(CFloatImage *img)
{
	app.ShowImage(img);
}

void FloatImage_Construct(UINT w, UINT h, CFloatImage *o)
{
	new(o) CFloatImage();
	o->Create(w,h);
}

void RegisterScriptImage(asIScriptEngine *engine)
{
	int r;
	r = engine->RegisterObjectType("image", sizeof(CFloatImage), asOBJ_CLASS_CDA); assert(r >= 0);
	r = engine->RegisterObjectBehaviour("image", asBEHAVE_CONSTRUCT, "void f()", asFUNCTIONPR(FloatImage_Construct, (CFloatImage*), void), asCALL_CDECL_OBJLAST); assert(r >= 0);
	r = engine->RegisterObjectBehaviour("image", asBEHAVE_CONSTRUCT, "void f(uint, uint)", asFUNCTIONPR(FloatImage_Construct, (asUINT, asUINT, CFloatImage*), void), asCALL_CDECL_OBJLAST); assert(r >= 0);
	r = engine->RegisterObjectBehaviour("image", asBEHAVE_ADDREF, "void f()", asMETHOD(CFloatImage,AddRef), asCALL_THISCALL); assert(r >= 0);
	r = engine->RegisterObjectBehaviour("image", asBEHAVE_RELEASE, "void f()", asMETHOD(CFloatImage,Release), asCALL_THISCALL); assert(r >= 0);
	r = engine->RegisterObjectBehaviour("image", asBEHAVE_ASSIGNMENT, "image &f(const image &in)", asMETHOD(CFloatImage,operator=), asCALL_THISCALL); assert(r >= 0);
	r = engine->RegisterObjectProperty("image", "const uint width", offsetof(CFloatImage, width)); assert(r >= 0);
	r = engine->RegisterObjectProperty("image", "const uint height", offsetof(CFloatImage, height)); assert(r >= 0);

	r = engine->RegisterObjectMethod("image", "void Resize(uint, uint)", asMETHOD(CFloatImage,Resize), asCALL_THISCALL); assert(r >= 0);
	r = engine->RegisterObjectMethod("image", "void Show()", asFUNCTION(FloatImage_Show), asCALL_CDECL_OBJLAST); assert(r >= 0);

	// Register pixel mutator methods
    r = engine->RegisterObjectMethod("image", "float &Alpha(uint x, uint y)", asMETHOD(CFloatImage,Alpha), asCALL_THISCALL); assert(r >= 0);
    r = engine->RegisterObjectMethod("image", "float &Red(uint x, uint y)", asMETHOD(CFloatImage,Red), asCALL_THISCALL); assert(r >= 0);
    r = engine->RegisterObjectMethod("image", "float &Green(uint x, uint y)", asMETHOD(CFloatImage,Green), asCALL_THISCALL); assert(r >= 0);
    r = engine->RegisterObjectMethod("image", "float &Blue(uint x, uint y)", asMETHOD(CFloatImage,Blue), asCALL_THISCALL); assert(r >= 0);
    r = engine->RegisterObjectMethod("image", "float &Channel(uint ch, uint x, uint y)", asMETHOD(CFloatImage,Channel), asCALL_THISCALL); assert(r >= 0);
    r = engine->RegisterObjectMethod("image", "pixel &Pixel(uint x, uint y)", asMETHOD(CFloatImage,Blue), asCALL_THISCALL); assert(r >= 0);

	// Register pixel inspector methods (const overloaded)
    r = engine->RegisterObjectMethod("image", "const float &Alpha(uint x, uint y) const", asMETHOD(CFloatImage,Alpha), asCALL_THISCALL); assert(r >= 0);
    r = engine->RegisterObjectMethod("image", "const float &Red(uint x, uint y) const", asMETHOD(CFloatImage,Red), asCALL_THISCALL); assert(r >= 0);
    r = engine->RegisterObjectMethod("image", "const float &Green(uint x, uint y) const", asMETHOD(CFloatImage,Green), asCALL_THISCALL); assert(r >= 0);
    r = engine->RegisterObjectMethod("image", "const float &Blue(uint x, uint y) const", asMETHOD(CFloatImage,Blue), asCALL_THISCALL); assert(r >= 0);
    r = engine->RegisterObjectMethod("image", "const float &Channel(uint ch, uint x, uint y) const", asMETHOD(CFloatImage,Channel), asCALL_THISCALL); assert(r >= 0);
    r = engine->RegisterObjectMethod("image", "const pixel &Pixel(uint x, uint y) const", asMETHOD(CFloatImage,Blue), asCALL_THISCALL); assert(r >= 0);
}